import logging

from twisted.internet.defer import inlineCallbacks

from peek_core_search._private.client.DeviceTupleProcessorActionProxy import (
    makeTupleActionProcessorProxy,
)
from peek_core_search._private.client.controller.FastKeywordController import (
    FastKeywordController,
)
from peek_plugin_base.PeekVortexUtil import peekServerName
from peek_plugin_base.client.PluginClientEntryHookABC import PluginClientEntryHookABC
from peek_core_search._private.PluginNames import searchActionProcessorName
from peek_core_search._private.PluginNames import searchFilt
from peek_core_search._private.PluginNames import searchObservableName
from peek_core_search._private.client.TupleDataObservable import (
    makeClientTupleDataObservableHandler,
)
from peek_core_search._private.client.controller.SearchIndexCacheController import (
    SearchIndexCacheController,
)
from peek_core_search._private.client.controller.SearchObjectCacheController import (
    SearchObjectCacheController,
)
from peek_core_search._private.client.handlers.SearchIndexCacheHandler import (
    SearchIndexCacheHandler,
)
from peek_core_search._private.client.handlers.SearchObjectCacheHandler import (
    SearchObjectCacheHandler,
)
from peek_core_search._private.storage.DeclarativeBase import loadStorageTuples
from peek_core_search._private.tuples import loadPrivateTuples
from peek_core_search.tuples import loadPublicTuples
from vortex.handler.TupleActionProcessorProxy import TupleActionProcessorProxy
from vortex.handler.TupleDataObservableProxyHandler import (
    TupleDataObservableProxyHandler,
)
from vortex.handler.TupleDataObserverClient import TupleDataObserverClient

logger = logging.getLogger(__name__)


class ClientEntryHook(PluginClientEntryHookABC):
    def __init__(self, *args, **kwargs):
        """" Constructor """
        # Call the base classes constructor
        PluginClientEntryHookABC.__init__(self, *args, **kwargs)

        #: Loaded Objects, This is a list of all objects created when we start
        self._loadedObjects = []

    def load(self) -> None:
        """Load

        This will be called when the plugin is loaded, just after the db is migrated.
        Place any custom initialiastion steps here.

        """

        loadStorageTuples()

        loadPrivateTuples()
        loadPublicTuples()

        logger.debug("Loaded")

    @inlineCallbacks
    def start(self):
        """Load

        This will be called when the plugin is loaded, just after the db is migrated.
        Place any custom initialiastion steps here.

        """
        # ----------------

        # Provide the devices access to the servers observable
        tupleDataObservableProxyHandler = TupleDataObservableProxyHandler(
            observableName=searchObservableName,
            proxyToVortexName=peekServerName,
            additionalFilt=searchFilt,
            observerName="Proxy to devices",
        )
        self._loadedObjects.append(tupleDataObservableProxyHandler)

        # ----------------
        #: This is an observer for us (the client) to use to observe data
        # from the server
        serverTupleObserver = TupleDataObserverClient(
            observableName=searchObservableName,
            destVortexName=peekServerName,
            additionalFilt=searchFilt,
            observerName="Data for client",
        )
        self._loadedObjects.append(serverTupleObserver)

        # ----------------
        # Search Index Cache Controller

        searchIndexCacheController = SearchIndexCacheController(self.platform.serviceId)
        self._loadedObjects.append(searchIndexCacheController)

        # This is the custom handler for the client
        searchIndexCacheHandler = SearchIndexCacheHandler(
            cacheController=searchIndexCacheController, clientId=self.platform.serviceId
        )
        self._loadedObjects.append(searchIndexCacheHandler)

        searchIndexCacheController.setCacheHandler(searchIndexCacheHandler)

        # ----------------
        # Search Object Cache Controller

        searchObjectCacheController = SearchObjectCacheController(
            self.platform.serviceId
        )
        self._loadedObjects.append(searchObjectCacheController)

        # This is the custom handler for the client
        searchObjectCacheHandler = SearchObjectCacheHandler(
            cacheController=searchObjectCacheController,
            clientId=self.platform.serviceId,
        )
        self._loadedObjects.append(searchObjectCacheHandler)

        searchObjectCacheController.setCacheHandler(searchObjectCacheHandler)

        # ----------------
        # Fast Keyword Controller

        fastKeywordController = FastKeywordController(
            searchObjectCacheController, searchIndexCacheController
        )
        self._loadedObjects.append(fastKeywordController)

        searchIndexCacheController.setFastKeywordController(fastKeywordController)

        # ----------------
        # Proxy actions back to the server, we don't process them at all
        tupleActionHandler = makeTupleActionProcessorProxy(fastKeywordController)
        self._loadedObjects.append(tupleActionHandler)

        # ----------------
        # Create the Tuple Observer
        makeClientTupleDataObservableHandler(
            tupleDataObservableProxyHandler,
            searchIndexCacheController,
            searchObjectCacheController,
        )

        # ----------------
        # Start the compiler controllers
        yield searchIndexCacheController.start()
        yield searchObjectCacheController.start()

        logger.debug("Started")

    def stop(self):
        """Stop

        This method is called by the platform to tell the peek app to shutdown and stop
        everything it's doing
        """
        # Shutdown and dereference all objects we constructed when we started
        while self._loadedObjects:
            self._loadedObjects.pop().shutdown()

        logger.debug("Stopped")

    def unload(self):
        """Unload

        This method is called after stop is called, to unload any last resources
        before the PLUGIN is unlinked from the platform

        """
        logger.debug("Unloaded")



import string
from typing import Set


def _splitFullTokens(keywordStr: str) -> Set[str]:
    if not keywordStr:
        return set()

    # Remove punctuation
    tokens = keywordStr.lower().split(" ")

    # Strip and Split words, filter out words less than three letters
    return set([w.strip() for w in tokens if 2 <= len(w)])


def splitFullKeywords(keywordStr: str) -> Set[str]:
    """Split Full Keywords

    This tokenizer function is used for search strings loaded into
    the fullKwPropertiesJson field of the SearchObject.

    This exists, because not all data is suitable for partial keywork search
    for example IDs like J00001223COMP

    """
    tokens = _splitFullTokens(keywordStr)
    tokens = ["^%s$" % t for t in tokens]

    return set(tokens)


def splitPartialKeywords(keywordStr: str) -> Set[str]:
    """Split Partial Keywords

    This tokenizer function is used for search strings loaded into the
    partialKwPropertiesJson field of the SearchObject.

    This will tokenize
    * all the smaller keywords as a full keyword,
    * all numbers as full keywords
    * all remaining items will be tokenized in a rolling three char method.

    :param keywordStr: The string to tokenize

    """
    # Strip and Split words, filter out words less than three letters
    tokens = _splitFullTokens(keywordStr)

    # Split the words up into tokens, this creates partial keyword search support
    tempTokens, tokens = tokens, []
    for token in tempTokens:
        # Partial tokenize the other words.
        for index in range(max(1, len(token) - 2)):
            tokens.append(("" if index else "^") + token[index : index + 3])

    return set(tokens)


def twoCharTokens(tokens: Set[str]) -> Set[str]:
    def isTwoChar(token: str) -> bool:
        if token.endswith("$") and len(token) == 4:
            return True
        if token.startswith("^") and len(token) == 3:
            return True
        return False

    return set([t for t in tokens if isTwoChar(t)])



from vortex.handler.TupleDataObservableProxyHandler import (
    TupleDataObservableProxyHandler,
)

from peek_core_search._private.client.controller.SearchIndexCacheController import (
    SearchIndexCacheController,
)
from peek_core_search._private.client.controller.SearchObjectCacheController import (
    SearchObjectCacheController,
)
from peek_core_search._private.client.tuple_providers.ClientSearchIndexUpdateDateTupleProvider import (
    ClientSearchIndexUpdateDateTupleProvider,
)
from peek_core_search._private.client.tuple_providers.ClientSearchObjectUpdateDateTupleProvider import (
    ClientSearchObjectUpdateDateTupleProvider,
)
from peek_core_search._private.tuples.search_index.SearchIndexUpdateDateTuple import (
    SearchIndexUpdateDateTuple,
)
from peek_core_search._private.tuples.search_object.SearchObjectUpdateDateTuple import (
    SearchObjectUpdateDateTuple,
)


def makeClientTupleDataObservableHandler(
    tupleObservable: TupleDataObservableProxyHandler,
    searchIndexCacheHandler: SearchIndexCacheController,
    searchObjectCacheHandler: SearchObjectCacheController,
):
    """ " Make CLIENT Tuple Data Observable Handler

    This method creates the observable object, registers the tuple providers and then
    returns it.

    :param tupleObservable: The tuple observable proxy
    :param searchIndexCacheHandler: The search index cache handler
    :param searchObjectCacheHandler: The search object cache handler
    :return: An instance of :code:`TupleDataObservableHandler`

    """

    tupleObservable.addTupleProvider(
        SearchIndexUpdateDateTuple.tupleName(),
        ClientSearchIndexUpdateDateTupleProvider(searchIndexCacheHandler),
    )

    tupleObservable.addTupleProvider(
        SearchObjectUpdateDateTuple.tupleName(),
        ClientSearchObjectUpdateDateTupleProvider(searchObjectCacheHandler),
    )



from twisted.internet.defer import inlineCallbacks
from twisted.trial import unittest

from peek_core_search._private.client.controller.FastKeywordController import (
    FastKeywordController,
)
from peek_core_search._private.tuples.search_object.SearchResultObjectTuple import (
    SearchResultObjectTuple,
)
from peek_core_search._private.worker.tasks.KeywordSplitter import splitPartialKeywords


class FastKeywordControllerTest(unittest.TestCase):
    def test_mergePartialAndFullMatches_1(self):
        fullByKw = {"^to$": [5, 6]}
        partialByKw = {"^to": [7], "^aa": [8]}

        inst = FastKeywordController(None, None)

        resultByKw = inst._mergePartialAndFullMatches("to", fullByKw, partialByKw)

        self.assertEqual(set(resultByKw), {"to"})
        self.assertEqual(len(resultByKw), 1)
        self.assertEqual(set(resultByKw["to"]), {5, 6, 7})

    def test_mergePartialAndFullMatches_2(self):
        fullByKw = {"^five$": [5, 6]}
        partialByKw = {"^fiv": [7, 6], "ive": [7, 6]}

        inst = FastKeywordController(None, None)

        resultByKw = inst._mergePartialAndFullMatches(
            "five seven", fullByKw, partialByKw
        )

        self.assertEqual(set(resultByKw), {"five"})
        self.assertEqual(len(resultByKw), 1)
        self.assertEqual(set(resultByKw["five"]), {5, 6, 7})

    def test_mergePartialAndFullMatches_3(self):
        searchString = "tatu west fus"
        fullByKw = {}
        partialByKw = {t: [7, 6] for t in splitPartialKeywords(searchString)}

        inst = FastKeywordController(None, None)

        resultByKw = inst._mergePartialAndFullMatches(
            searchString, fullByKw, partialByKw
        )

        self.assertEqual(set(resultByKw), {"west", "fus", "tatu"})
        self.assertEqual(set(resultByKw["west"]), {6, 7})
        self.assertEqual(set(resultByKw["fus"]), {6, 7})
        self.assertEqual(set(resultByKw["tatu"]), {6, 7})

    def test_setIntersectFilterIndexResults(self):
        data = {"0": [6778979, 7042955]}

        inst = FastKeywordController(None, None)

        result = inst._setIntersectFilterIndexResults(data)

        self.assertEqual(result, {6778979, 7042955})
        self.assertEqual(len(result), len(data["0"]))

    @inlineCallbacks
    def test_filterObjectsForSearchString(self):
        from twisted.python import threadable

        threadable.isInIOThread = lambda: True

        data = [
            SearchResultObjectTuple(properties={"any": "TRANS HV FUSE TATURA WEST 28"})
        ]

        inst = FastKeywordController(None, None)

        # No property name
        result = yield inst._filterAndRankObjectsForSearchString(
            data, "TRANS HV FUSE TATURA WEST 28", None
        )

        self.assertEqual(len(result), 1)

        # Now try it with the property name
        result = yield inst._filterAndRankObjectsForSearchString(
            data, "TRANS HV FUSE TATURA WEST 28", "any"
        )

        self.assertEqual(len(result), 1)

        # Now try it with a WRONG the property name
        result = yield inst._filterAndRankObjectsForSearchString(
            data, "TRANS HV FUSE TATURA WEST 28", "dfgdfg"
        )

        self.assertEqual(len(result), 0)

        # partial keywords search string
        result = yield inst._filterAndRankObjectsForSearchString(data, "tatu west fus", None)

        self.assertEqual(len(result), 1)



"""Increased chunk size

Peek Plugin Database Migration Script

Revision ID: c18d0a2ab678
Revises: 2c6cad1f280e
Create Date: 2018-07-04 21:56:28.319589

"""

# revision identifiers, used by Alembic.

from sqlalchemy import MetaData, Integer, String, Column
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from peek_core_search._private.worker.tasks._CalcChunkKey import (
    makeSearchIndexChunkKey,
    makeSearchObjectChunkKey,
)

revision = "c18d0a2ab678"
down_revision = "2c6cad1f280e"
branch_labels = None
depends_on = None

from alembic import op

__DeclarativeBase = declarative_base(metadata=MetaData(schema="core_search"))

import logging

logger = logging.getLogger(__name__)


class __SearchIndex(__DeclarativeBase):
    __tablename__ = "SearchIndex"

    id = Column(Integer, primary_key=True, autoincrement=True)

    chunkKey = Column(Integer, nullable=False)
    keyword = Column(String, nullable=False)


class __SearchObject(__DeclarativeBase):
    __tablename__ = "SearchObject"

    id = Column(Integer, primary_key=True, autoincrement=True)
    key = Column(String, nullable=False)
    chunkKey = Column(Integer, nullable=False)


def _loadSearchObjects(session, _Declarative):
    FETCH_SIZE = 5000
    lastOffset = 0
    while True:
        rows = (
            session.query(_Declarative)
            .order_by(_Declarative.id)
            .offset(lastOffset)
            .limit(FETCH_SIZE)
            .yield_per(FETCH_SIZE)
            .all()
        )
        if not rows:
            return
        logger.info(
            "Loading %s-%s for %s",
            lastOffset,
            lastOffset + FETCH_SIZE,
            _Declarative.__name__,
        )
        yield rows
        lastOffset += FETCH_SIZE


def upgrade():
    bind = op.get_bind()
    session = sessionmaker()(bind=bind)

    for rows in _loadSearchObjects(session, __SearchIndex):
        for item in rows:
            item.chunkKey = makeSearchIndexChunkKey(item.keyword)
        session.commit()
        session.expunge_all()

    for rows in _loadSearchObjects(session, __SearchObject):
        for item in rows:
            item.chunkKey = makeSearchObjectChunkKey(item.id)
        session.commit()
        session.expunge_all()

    session.close()

    op.execute(' TRUNCATE TABLE core_search."EncodedSearchIndexChunk" ')
    op.execute(' TRUNCATE TABLE core_search."SearchIndexCompilerQueue" ')

    op.execute(
        """INSERT INTO core_search."SearchIndexCompilerQueue"
                            ("chunkKey")
                            SELECT DISTINCT "chunkKey"
                            FROM core_search."SearchIndex"
                         """
    )

    op.execute(' TRUNCATE TABLE core_search."EncodedSearchObjectChunk" ')
    op.execute(' TRUNCATE TABLE core_search."SearchObjectCompilerQueue" ')

    op.execute(
        """INSERT INTO core_search."SearchObjectCompilerQueue"
                            ("chunkKey")
                            SELECT DISTINCT "chunkKey"
                            FROM core_search."SearchObject"
                         """
    )


def downgrade():
    raise NotImplementedError("Downgrade not implemented")



import json
import logging
from collections import defaultdict
from datetime import datetime
from typing import List, Dict, Tuple, Set

import pytz
from peek_core_search._private.storage.SearchObject import SearchObject
from peek_core_search._private.storage.SearchObjectCompilerQueue import (
    SearchObjectCompilerQueue,
)
from peek_core_search._private.storage.SearchObjectRoute import SearchObjectRoute
from peek_core_search._private.storage.SearchObjectTypeTuple import (
    SearchObjectTypeTuple,
)
from peek_core_search._private.storage.SearchPropertyTuple import SearchPropertyTuple
from peek_core_search._private.worker.tasks.ImportSearchIndexTask import (
    ObjectToIndexTuple,
    reindexSearchObject,
)
from peek_core_search._private.worker.tasks._CalcChunkKey import (
    makeSearchObjectChunkKey,
)
from peek_core_search.tuples.ImportSearchObjectTuple import ImportSearchObjectTuple
from peek_plugin_base.worker import CeleryDbConn
from peek_plugin_base.worker.CeleryApp import celeryApp
from sqlalchemy import select, bindparam, and_, or_
from txcelery.defer import DeferrableTask
from vortex.Payload import Payload

logger = logging.getLogger(__name__)


# We need to insert the into the following tables:
# SearchObject - or update it's details if required
# SearchIndex - The index of the keywords for the object
# SearchObjectRoute - delete old importGroupHash
# SearchObjectRoute - insert the new routes


@DeferrableTask
@celeryApp.task(bind=True)
def removeSearchObjectTask(self, importGroupHashes: List[str]) -> None:
    pass


@DeferrableTask
@celeryApp.task(bind=True)
def importSearchObjectTask(self, searchObjectsEncodedPayload: bytes) -> None:
    startTime = datetime.now(pytz.utc)

    # Decode arguments
    newSearchObjects: List[ImportSearchObjectTuple] = (
        Payload().fromEncodedPayload(searchObjectsEncodedPayload).tuples
    )

    # Cleanup some of the input data
    for o in newSearchObjects:
        if not o.objectType:
            o.objectType = "none"

        o.objectType = o.objectType.lower()

        if o.fullKeywords:
            o.fullKeywords = {k.lower(): v for k, v in o.fullKeywords.items()}

        if o.partialKeywords:
            o.partialKeywords = {k.lower(): v for k, v in o.partialKeywords.items()}

    try:
        objectTypeIdsByName = _prepareLookups(newSearchObjects)

        objectIdByKey, chunkKeysForQueue = _insertOrUpdateObjects(
            newSearchObjects, objectTypeIdsByName
        )

        _insertObjectRoutes(newSearchObjects, objectIdByKey)

        _packObjectJson(list(objectIdByKey.values()), chunkKeysForQueue)

        logger.info(
            "Imported %s SearchObjects in %s",
            len(newSearchObjects),
            datetime.now(pytz.utc) - startTime,
        )

    except Exception as e:
        logger.debug("Retrying import search objects, %s", e)
        raise self.retry(exc=e, countdown=3)


def _prepareLookups(newSearchObjects: List[ImportSearchObjectTuple]) -> Dict[str, int]:
    """Check Or Insert Search Properties

    Make sure the search properties exist.

    """

    dbSession = CeleryDbConn.getDbSession()

    startTime = datetime.now(pytz.utc)

    try:

        objectTypeNames = {"none"}
        propertyNames = {"key"}

        for o in newSearchObjects:
            objectTypeNames.add(o.objectType)

            if o.fullKeywords:
                propertyNames.update(o.fullKeywords)

            if o.partialKeywords:
                propertyNames.update(o.partialKeywords)

        # Prepare Properties
        dbProps = dbSession.query(SearchPropertyTuple).all()
        propertyNames -= set([o.name for o in dbProps])

        if propertyNames:
            for newPropName in propertyNames:
                dbSession.add(SearchPropertyTuple(name=newPropName, title=newPropName))

            dbSession.commit()

        del dbProps
        del propertyNames

        # Prepare Object Types
        dbObjectTypes = dbSession.query(SearchObjectTypeTuple).all()
        objectTypeNames -= set([o.name for o in dbObjectTypes])

        if not objectTypeNames:
            objectTypeIdsByName = {o.name: o.id for o in dbObjectTypes}

        else:
            for newPropName in objectTypeNames:
                dbSession.add(
                    SearchObjectTypeTuple(name=newPropName, title=newPropName)
                )

            dbSession.commit()

            dbObjectTypes = dbSession.query(SearchObjectTypeTuple).all()
            objectTypeIdsByName = {o.name: o.id for o in dbObjectTypes}

        logger.debug("Prepared lookups in %s", (datetime.now(pytz.utc) - startTime))

        return objectTypeIdsByName

    except Exception as e:
        dbSession.rollback()
        raise

    finally:
        dbSession.close()


def _insertOrUpdateObjects(
    newSearchObjects: List[ImportSearchObjectTuple], objectTypeIdsByName: Dict[str, int]
) -> Tuple[Dict[str, int], Set[int]]:
    """Insert or Update Objects

    1) Find objects and update them
    2) Insert object if the are missing

    """

    searchObjectTable = SearchObject.__table__

    startTime = datetime.now(pytz.utc)

    createdObjectByKey, engine, newIdGen = _loadExistingObjects(
        newSearchObjects, searchObjectTable
    )

    conn = engine.connect()
    transaction = conn.begin()

    try:

        # Create state arrays
        objectsToIndex: Dict[int, ObjectToIndexTuple] = {}
        objectIdByKey: Dict[str, int] = {}
        inserts = []
        propUpdates = []
        objectTypeUpdates = []
        chunkKeysForQueue: Set[int] = set()

        # Work out which objects have been updated or need inserting
        for importObject in newSearchObjects:
            originalImportObjectKey = importObject.key
            importObject.key = importObject.key.lower()
            loweredObjectKey = importObject.key

            existingObject = createdObjectByKey.get(loweredObjectKey)
            importObjectTypeId = objectTypeIdsByName[importObject.objectType]

            fullKwPropsWithKey = dict(key=originalImportObjectKey)
            partialKwProps = {}

            if importObject.fullKeywords:
                if existingObject and existingObject.fullKwPropertiesJson:
                    fullKwPropsWithKey.update(
                        json.loads(existingObject.fullKwPropertiesJson)
                    )

                # Add the data we're importing second
                # Remove null values
                fullKwPropsWithKey.update(importObject.fullKeywords)
                fullKwPropsWithKey = {k: v for k, v in fullKwPropsWithKey.items() if v}

            if importObject.partialKeywords:
                if existingObject and existingObject.partialKwPropertiesJson:
                    partialKwProps.update(
                        json.loads(existingObject.partialKwPropertiesJson)
                    )

                # Add the data we're importing second
                # Remove null values
                partialKwProps.update(importObject.partialKeywords)
                partialKwProps = {
                    k: v for k, v in importObject.partialKeywords.items() if v
                }

            fullKwPropsStr = json.dumps(fullKwPropsWithKey, sort_keys=True)
            partialKwPropsStr = json.dumps(partialKwProps, sort_keys=True)

            # Work out if we need to update the object type
            if importObject.objectType != "None" and existingObject:
                objectTypeUpdates.append(
                    dict(b_id=existingObject.id, b_typeId=importObjectTypeId)
                )

            # Work out if we need to update the existing object or create one
            if existingObject:
                searchIndexUpdateNeeded = (
                    fullKwPropsStr
                    and existingObject.fullKwPropertiesJson != fullKwPropsStr
                )

                searchIndexUpdateNeeded = (
                    searchIndexUpdateNeeded
                    or partialKwPropsStr
                    and existingObject.partialKwPropertiesJson != partialKwPropsStr
                )

                if searchIndexUpdateNeeded:
                    propUpdates.append(
                        dict(
                            b_id=existingObject.id,
                            b_fullPropsKwStr=fullKwPropsStr,
                            b_partialKwPropsStr=partialKwPropsStr,
                        )
                    )

            else:
                searchIndexUpdateNeeded = True
                id_ = next(newIdGen)
                existingObject = SearchObject(
                    id=id_,
                    key=originalImportObjectKey,
                    objectTypeId=importObjectTypeId,
                    fullKwPropertiesJson=fullKwPropsStr,
                    partialKwPropertiesJson=partialKwPropsStr,
                    chunkKey=makeSearchObjectChunkKey(id_),
                )
                inserts.append(existingObject.tupleToSqlaBulkInsertDict())

                # Add our made up object to the created list, the logic of this loop
                # will merge and subsequent objects if they contain prop updates, etc
                createdObjectByKey[loweredObjectKey] = existingObject

            if searchIndexUpdateNeeded:
                objectsToIndex[existingObject.id] = ObjectToIndexTuple(
                    id=existingObject.id,
                    fullKwProps=fullKwPropsWithKey,
                    partialKwProps=partialKwProps,
                )

            objectIdByKey[loweredObjectKey] = existingObject.id
            chunkKeysForQueue.add(existingObject.chunkKey)

        # Insert the Search Objects
        if inserts:
            conn.execute(searchObjectTable.insert(), inserts)

        if propUpdates:
            stmt = (
                searchObjectTable.update()
                .where(searchObjectTable.c.id == bindparam("b_id"))
                .values(
                    fullKwPropertiesJson=bindparam("b_fullPropsKwStr"),
                    partialKwPropertiesJson=bindparam("b_partialKwPropsStr"),
                )
            )
            conn.execute(stmt, propUpdates)

        if objectTypeUpdates:
            stmt = (
                searchObjectTable.update()
                .where(searchObjectTable.c.id == bindparam("b_id"))
                .values(objectTypeId=bindparam("b_typeId"))
            )
            conn.execute(stmt, objectTypeUpdates)

        # Reindex the keywords
        reindexSearchObject(conn, list(objectsToIndex.values()))

        if objectsToIndex or inserts or propUpdates or objectTypeUpdates:
            transaction.commit()
        else:
            transaction.rollback()

        logger.debug(
            "Inserted %s updated %s ObjectToIndexTuple in %s",
            len(inserts),
            len(propUpdates),
            (datetime.now(pytz.utc) - startTime),
        )

        logger.debug("Passing to index %s SearchIndex", len(objectsToIndex))

        return objectIdByKey, chunkKeysForQueue

    except Exception as e:
        transaction.rollback()
        raise

    finally:
        conn.close()


def _loadExistingObjects(newSearchObjects, searchObjectTable):
    engine = CeleryDbConn.getDbEngine()
    conn = engine.connect()
    try:
        objectKeys = set(
            [o.key for o in newSearchObjects]
            + [o.key.lower() for o in newSearchObjects]
        )

        # Query existing objects
        results = list(
            conn.execute(
                select(
                    columns=[
                        searchObjectTable.c.id,
                        searchObjectTable.c.key,
                        searchObjectTable.c.chunkKey,
                        searchObjectTable.c.fullKwPropertiesJson,
                        searchObjectTable.c.partialKwPropertiesJson,
                    ],
                    whereclause=searchObjectTable.c.key.in_(objectKeys),
                )
            )
        )

        createdObjectByKey = {o.key.lower(): o for o in results}
        del results
        del objectKeys

    finally:
        conn.close()
    # Get the IDs that we need
    newSearchObjectUniqueCount = len(set([o.key.lower() for o in newSearchObjects]))
    newIdGen = CeleryDbConn.prefetchDeclarativeIds(
        SearchObject, newSearchObjectUniqueCount - len(createdObjectByKey)
    )
    del newSearchObjectUniqueCount
    return createdObjectByKey, engine, newIdGen


def _insertObjectRoutes(
    newSearchObjects: List[ImportSearchObjectTuple], objectIdByKey: Dict[str, int]
):
    """Insert Object Routes

    1) Drop all routes with matching importGroupHash

    2) Insert the new routes

    :param newSearchObjects:
    :param objectIdByKey:
    :return:
    """

    searchObjectRoute = SearchObjectRoute.__table__

    startTime = datetime.now(pytz.utc)

    engine = CeleryDbConn.getDbEngine()
    conn = engine.connect()
    transaction = conn.begin()

    try:
        importHashSet = set()
        inserts = []
        objectIdSet = set()

        # Make some lists to work out the existing routes
        for importObject in newSearchObjects:
            for impRoute in importObject.routes:
                importHashSet.add(impRoute.importGroupHash)
                objectIdSet.add(objectIdByKey[importObject.key])

        # Query for existing routes
        results = list(
            conn.execute(
                select(
                    columns=[
                        searchObjectRoute.c.objectId,
                        searchObjectRoute.c.routeTitle,
                        searchObjectRoute.c.importGroupHash,
                    ],
                    whereclause=and_(
                        searchObjectRoute.c.objectId.in_(objectIdSet),
                        ~searchObjectRoute.c.importGroupHash.in_(importHashSet),
                    ),
                )
            )
        )

        existingRoutes = {
            "%s.%s" % (o.objectId, o.routeTitle): dict(o) for o in results
        }
        newRoutes = {}
        del results

        # Now create the inserts
        for importObject in newSearchObjects:
            for impRoute in importObject.routes:
                objectId = objectIdByKey[importObject.key]
                routeInsert = dict(
                    objectId=objectId,
                    importGroupHash=impRoute.importGroupHash,
                    routeTitle=impRoute.routeTitle,
                    routePath=impRoute.routePath,
                )

                uniqueRouteStr = "%s.%s" % (objectId, impRoute.routeTitle)

                if uniqueRouteStr in existingRoutes:
                    logger.debug(
                        "A duplicate route exists in another" " import group\n%s\n%s",
                        existingRoutes[uniqueRouteStr],
                        routeInsert,
                    )

                elif uniqueRouteStr in newRoutes:
                    logger.debug(
                        "Duplicate route titles defined in this"
                        " import group\n%s\n%s",
                        newRoutes[uniqueRouteStr],
                        routeInsert,
                    )

                else:
                    inserts.append(routeInsert)
                    newRoutes[uniqueRouteStr] = routeInsert

        if importHashSet:
            conn.execute(
                searchObjectRoute.delete(
                    searchObjectRoute.c.importGroupHash.in_(importHashSet)
                )
            )

        # Insert the Search Object routes
        if inserts:
            conn.execute(searchObjectRoute.insert(), inserts)

        if importHashSet or inserts:
            transaction.commit()
        else:
            transaction.rollback()

        logger.debug(
            "Inserted %s SearchObjectRoute in %s",
            len(inserts),
            (datetime.now(pytz.utc) - startTime),
        )

    except Exception as e:
        transaction.rollback()
        raise

    finally:
        conn.close()


def _packObjectJson(updatedIds: List[int], chunkKeysForQueue: Set[int]):
    """Pack Object Json

    1) Create JSON and update object.

    Doing this takes longer to bulk load, but quicker to make incremental changes

    :param updatedIds:
    :param chunkKeysForQueue:
    :return:
    """

    searchObjectTable = SearchObject.__table__
    objectQueueTable = SearchObjectCompilerQueue.__table__
    dbSession = CeleryDbConn.getDbSession()

    startTime = datetime.now(pytz.utc)

    try:

        indexQry = (
            dbSession.query(
                SearchObject.id,
                SearchObject.fullKwPropertiesJson,
                SearchObject.partialKwPropertiesJson,
                SearchObject.objectTypeId,
                SearchObjectRoute.routeTitle,
                SearchObjectRoute.routePath,
            )
            .outerjoin(SearchObjectRoute, SearchObject.id == SearchObjectRoute.objectId)
            .filter(SearchObject.id.in_(updatedIds))
            .filter(
                or_(
                    SearchObject.fullKwPropertiesJson != None,
                    SearchObject.partialKwPropertiesJson != None,
                )
            )
            .order_by(SearchObject.id, SearchObjectRoute.routeTitle)
            .yield_per(1000)
            .all()
        )

        # I chose a simple name for this one.
        qryDict = defaultdict(list)

        for item in indexQry:
            (
                qryDict[
                    (
                        item.id,
                        item.fullKwPropertiesJson,
                        item.partialKwPropertiesJson,
                        item.objectTypeId,
                    )
                ].append([item.routeTitle, item.routePath])
            )

        packedJsonUpdates = []

        # Sort each bucket by the key
        for (
            id_,
            fullKwPropJson,
            partialKwPropJson,
            objectTypeId,
        ), routes in qryDict.items():
            props = json.loads(fullKwPropJson)
            props.update(json.loads(partialKwPropJson))
            props["_r_"] = routes
            props["_otid_"] = objectTypeId
            packedJson = json.dumps(props, sort_keys=True)
            packedJsonUpdates.append(dict(b_id=id_, b_packedJson=packedJson))

        if packedJsonUpdates:
            stmt = (
                searchObjectTable.update()
                .where(searchObjectTable.c.id == bindparam("b_id"))
                .values(packedJson=bindparam("b_packedJson"))
            )
            dbSession.execute(stmt, packedJsonUpdates)

        if chunkKeysForQueue:
            dbSession.execute(
                objectQueueTable.insert(), [dict(chunkKey=v) for v in chunkKeysForQueue]
            )

        if packedJsonUpdates or chunkKeysForQueue:
            dbSession.commit()
        else:
            dbSession.rollback()

        logger.debug(
            "Packed JSON for %s SearchObjects in %s",
            len(updatedIds),
            (datetime.now(pytz.utc) - startTime),
        )

    except Exception as e:
        dbSession.rollback()
        raise

    finally:
        dbSession.close()
